package org.eclipse.jdt.internal.compiler.parser.diagnose;

import org.eclipse.jdt.core.compiler.InvalidInputException;
import org.eclipse.jdt.internal.compiler.util.Util;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.parser.Scanner;
import org.eclipse.jdt.internal.compiler.parser.TerminalTokens;

public class LexStream implements TerminalTokens
{
    public static final int IS_AFTER_JUMP = 1;
    public static final int LBRACE_MISSING = 2;
    private int tokenCacheIndex;
    private int tokenCacheEOFIndex;
    private LexStream.Token[] tokenCache;
    private int currentIndex;
    private Scanner scanner;
    private int[] intervalStartToSkip;
    private int[] intervalEndToSkip;
    private int[] intervalFlagsToSkip;
    private int previousInterval;
    private int currentInterval;
    private boolean awaitingColonColon;
    
    public LexStream(int size, Scanner scanner, int[] intervalStartToSkip, int[] intervalEndToSkip, int[] intervalFlagsToSkip, int firstToken, int init, int eof) {
        this.currentIndex = -1;
        this.previousInterval = -1;
        this.currentInterval = -1;
        this.tokenCache = new LexStream.Token[size];
        this.tokenCacheIndex = 0;
        this.tokenCacheEOFIndex = Integer.MAX_VALUE;
        this.tokenCache[0] = new LexStream.Token();
        this.tokenCache[0].kind = firstToken;
        this.tokenCache[0].name = CharOperation.NO_CHAR;
        this.tokenCache[0].start = init;
        this.tokenCache[0].end = init;
        this.tokenCache[0].line = 0;
        this.intervalStartToSkip = intervalStartToSkip;
        this.intervalEndToSkip = intervalEndToSkip;
        this.intervalFlagsToSkip = intervalFlagsToSkip;
        this.awaitingColonColon = false;
        scanner.resetTo(init, eof);
        this.scanner = scanner;
    }
    
    private void readTokenFromScanner() {
        int length = this.tokenCache.length;
        boolean tokenNotFound = true;
        while (tokenNotFound) {
            try {
                int tokenKind = this.scanner.getNextToken();
                if (tokenKind == 83) {
                    this.awaitingColonColon = true;
                }
                else if (tokenKind == 7) {
                    this.awaitingColonColon = false;
                }
                if (tokenKind != 60) {
                    int start = this.scanner.getCurrentTokenStartPosition();
                    int end = this.scanner.getCurrentTokenEndPosition();
                    int nextInterval = this.currentInterval + 1;
                    if (this.intervalStartToSkip.length == 0 || nextInterval >= this.intervalStartToSkip.length || start < this.intervalStartToSkip[nextInterval]) {
                        LexStream.Token token2 = new LexStream.Token();
                        token2.kind = tokenKind;
                        token2.name = this.scanner.getCurrentTokenSource();
                        token2.start = start;
                        token2.end = end;
                        token2.line = Util.getLineNumber(end, this.scanner.lineEnds, 0, this.scanner.linePtr);
                        if (this.currentInterval != this.previousInterval && (this.intervalFlagsToSkip[this.currentInterval] & 0x2) == 0x0) {
                            token2.flags = 1;
                            if ((this.intervalFlagsToSkip[this.currentInterval] & 0x1) != 0x0) {
                                token2.flags |= 0x2;
                            }
                        }
                        this.previousInterval = this.currentInterval;
                        LexStream.Token[] tokenCache2 = this.tokenCache;
                        int tokenCacheIndex2 = this.tokenCacheIndex + 1;
                        this.tokenCacheIndex = tokenCacheIndex2;
                        tokenCache2[tokenCacheIndex2 % length] = token2;
                        tokenNotFound = false;
                    }
                    else {
                        Scanner scanner = this.scanner;
                        int[] intervalEndToSkip = this.intervalEndToSkip;
                        int currentInterval = this.currentInterval + 1;
                        this.currentInterval = currentInterval;
                        scanner.resetTo(intervalEndToSkip[currentInterval] + 1, this.scanner.eofPosition - 1);
                    }
                }
                else {
                    int start = this.scanner.getCurrentTokenStartPosition();
                    int end = this.scanner.getCurrentTokenEndPosition();
                    LexStream.Token token = new LexStream.Token();
                    token.kind = tokenKind;
                    token.name = CharOperation.NO_CHAR;
                    token.start = start;
                    token.end = end;
                    token.line = Util.getLineNumber(end, this.scanner.lineEnds, 0, this.scanner.linePtr);
                    LexStream.Token[] tokenCache = this.tokenCache;
                    int tokenCacheIndex = this.tokenCacheIndex + 1;
                    this.tokenCacheIndex = tokenCacheIndex;
                    tokenCache[tokenCacheIndex % length] = token;
                    this.tokenCacheEOFIndex = this.tokenCacheIndex;
                    tokenNotFound = false;
                }
            }
            catch (InvalidInputException ex) {}
        }
    }
    
    public LexStream.Token token(int index) {
        if (index < 0) {
            LexStream.Token eofToken = new LexStream.Token();
            eofToken.kind = 60;
            eofToken.name = CharOperation.NO_CHAR;
            return eofToken;
        }
        if (this.tokenCacheEOFIndex >= 0 && index > this.tokenCacheEOFIndex) {
            return this.token(this.tokenCacheEOFIndex);
        }
        int length = this.tokenCache.length;
        if (index > this.tokenCacheIndex) {
            int tokensToRead = index - this.tokenCacheIndex;
            while (tokensToRead-- != 0) {
                this.readTokenFromScanner();
            }
        }
        else if (this.tokenCacheIndex - length >= index) {
            return null;
        }
        return this.tokenCache[index % length];
    }
    
    public int getToken() {
        return this.currentIndex = this.next(this.currentIndex);
    }
    
    public int previous(int tokenIndex) {
        return (tokenIndex > 0) ? (tokenIndex - 1) : 0;
    }
    
    public int next(int tokenIndex) {
        return (tokenIndex < this.tokenCacheEOFIndex) ? (tokenIndex + 1) : this.tokenCacheEOFIndex;
    }
    
    public boolean afterEol(int i) {
        return i < 1 || this.line(i - 1) < this.line(i);
    }
    
    public void reset() {
        this.currentIndex = -1;
    }
    
    public void reset(int i) {
        this.currentIndex = this.previous(i);
    }
    
    public int badtoken() {
        return 0;
    }
    
    public int kind(int tokenIndex) {
        return this.token(tokenIndex).kind;
    }
    
    public char[] name(int tokenIndex) {
        return this.token(tokenIndex).name;
    }
    
    public int line(int tokenIndex) {
        return this.token(tokenIndex).line;
    }
    
    public int start(int tokenIndex) {
        return this.token(tokenIndex).start;
    }
    
    public int end(int tokenIndex) {
        return this.token(tokenIndex).end;
    }
    
    public int flags(int tokenIndex) {
        return this.token(tokenIndex).flags;
    }
    
    public boolean isInsideStream(int index) {
        return (this.tokenCacheEOFIndex < 0 || index <= this.tokenCacheEOFIndex) && (index > this.tokenCacheIndex || this.tokenCacheIndex - this.tokenCache.length < index);
    }
    
    @Override
    public String toString() {
        StringBuffer res = new StringBuffer();
        String source = new String(this.scanner.source);
        if (this.currentIndex < 0) {
            int previousEnd2 = -1;
            int intervalEnd2;
            for (int j = 0; j < this.intervalStartToSkip.length; previousEnd2 = intervalEnd2, ++j) {
                int intervalStart2 = this.intervalStartToSkip[j];
                intervalEnd2 = this.intervalEndToSkip[j];
                res.append(source.substring(previousEnd2 + 1, intervalStart2));
                res.append('<');
                res.append('@');
                res.append(source.substring(intervalStart2, intervalEnd2 + 1));
                res.append('@');
                res.append('>');
            }
            res.append(source.substring(previousEnd2 + 1));
        }
        else {
            LexStream.Token token = this.token(this.currentIndex);
            int curtokKind = token.kind;
            int curtokStart = token.start;
            int curtokEnd = token.end;
            int previousEnd = -1;
            int intervalEnd;
            for (int i = 0; i < this.intervalStartToSkip.length; previousEnd = intervalEnd, ++i) {
                int intervalStart = this.intervalStartToSkip[i];
                intervalEnd = this.intervalEndToSkip[i];
                if (curtokStart >= previousEnd && curtokEnd <= intervalStart) {
                    res.append(source.substring(previousEnd + 1, curtokStart));
                    res.append('<');
                    res.append('#');
                    res.append(source.substring(curtokStart, curtokEnd + 1));
                    res.append('#');
                    res.append('>');
                    res.append(source.substring(curtokEnd + 1, intervalStart));
                }
                else {
                    res.append(source.substring(previousEnd + 1, intervalStart));
                }
                res.append('<');
                res.append('@');
                res.append(source.substring(intervalStart, intervalEnd + 1));
                res.append('@');
                res.append('>');
            }
            if (curtokStart >= previousEnd) {
                res.append(source.substring(previousEnd + 1, curtokStart));
                res.append('<');
                res.append('#');
                if (curtokKind == 60) {
                    res.append("EOF#>");
                }
                else {
                    res.append(source.substring(curtokStart, curtokEnd + 1));
                    res.append('#');
                    res.append('>');
                    res.append(source.substring(curtokEnd + 1));
                }
            }
            else {
                res.append(source.substring(previousEnd + 1));
            }
        }
        return res.toString();
    }
    
    public boolean awaitingColonColon() {
        return this.awaitingColonColon;
    }
    
    public static class Token
    {
        int kind;
        char[] name;
        int start;
        int end;
        int line;
        int flags;
        
        @Override
        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append(this.name).append('[').append(this.kind).append(']');
            buffer.append('{').append(this.start).append(',').append(this.end).append('}').append(this.line);
            return buffer.toString();
        }
    }
}
